﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module common.connectiontable;

// コネクションテーブル

private import std.c.time;
private import std.socket;
private import pes.net.winny.all;
private import pes.net.util;
private import pes.system.all;
//private import pes.prompt;
private import common.nodetable;
private import common.configuration;

// コネクション要素
class Connection
{
	ulong				debug_key;
protected:
	NyConnection		m_connection;
	time_t				m_time;
	NodeDetails			m_node;
	bool				m_lock;
	time_t				m_last_diffusion_response_time;
	time_t				m_last_diffusion_request_time;
	
public:
	invariant {
		if (!(debug_key == 0 || debug_key == m_node.aaKey)) {
			writef("Assert--C- %s:%s\n", debug_key, m_node.aaKey);
			assert(0);
		}
	}
	
	// 能動: 上流検索リンク用
	// 新規接続する
	// テーブルに入れるかはその後の話
	NyProtocolHeader connect(Configuration config, NodeDetails node,
							NyConnectionType.LinkType link_type = NyConnectionType.LinkType.SEARCH,
							uint timeout_sec = uint.max)
	{
		NyProtocolHeader 	header;
		NySpeed				speed;
		NyConnectionType	type;
		NyNodeDetails		details;
		
		close();	// 一度クローズしとく
		
		m_connection = new NyConnection;
		
		if (timeout_sec == uint.max) {
			m_connection.timeout(config.timeout, 0);
		} else {
			m_connection.timeout(timeout_sec, 0);
		}
		
		// negotiation
		try {
			header = m_connection.connect(node.inetAddress);
			
//			writef("---- %s\n", header.toString);
			
			sendNodeDetails(m_connection, config, link_type);
			receiveNodeDetails(m_connection, speed, type, details);
			node.know(config, node.inetAddress, header, speed, type, details);
		} catch (Exception e) {
			close();
			// クローズだけするので例外は上でキャッチする!!
			// つまりコネクトに失敗した場合は、クローズされている!!
			throw e; // 再送
		}
		
		m_node 	= node;
		m_time	= time(null);
		updateLastDiffusionRequestTime(m_time + 5);
		
		return header;
	}
	
	bool connectTest(Configuration config, InternetAddress addr)
	{
		NyProtocolHeader 	header;
		NySpeed				speed;
		NyConnectionType	type;
		NyNodeDetails		details;
		
		close();	// 一度クローズしとく
		
		m_connection = new NyConnection;
		
		m_connection.timeout(config.timeout, 0);
		
		// negotiation
		try {
			header = m_connection.connect(addr);
			sendNodeDetails(m_connection, config, NyConnectionType.LinkType.TRANSFER, true);
			receiveNodeDetails(m_connection, speed, type, details);
		} catch (Exception e) {
			return false;
		} finally {
			close();
		}
		
		return true;
	}
	
	// 受動: 下流検索リンク、転送リンク用
	// connは、コネクトしてきた接続を渡す
	NodeDetails accept(Configuration config, NyConnection conn,
							NyProtocolHeader header, out NyConnectionType type)
	{
		NySpeed			speed;
		NyNodeDetails	details;
		
		close();	// 一度クローズしとく
		
		NodeDetails node = new NodeDetails(conn.address);
		
		m_connection 	= conn;
		m_connection.timeout(config.timeout, 0);
		
		try {
			receiveNodeDetails(m_connection, speed, type, details);
			sendNodeDetails(m_connection, config, type.linkType);// 同じタイプを返してみる
			node.know(config, node.inetAddress, header, speed, type, details);
		} catch (Exception e) {
			close();
			// クローズだけするので例外は上でキャッチする!!
			// つまりにアクセプト失敗した場合は、クローズされている!!
			throw e;
		}
		
		m_node			= node;
		m_time			= time(null);
		updateLastDiffusionRequestTime(m_time + 5);
		
		return node;
	}
	
	void close(NyCommand close_command = null)
	{
//		m_node = null;
		
		if (m_connection !is null) {
			try {
				if (close_command is null) {
					close_command = new NyClose;
				}
				m_connection.sendCommand(close_command);
			} finally {
				m_connection.close();
//				m_connection = null;
			}
		}
	}
	
	// ノード情報送信
	static void sendNodeDetails(NyConnection conn, Configuration config,
										NyConnectionType.LinkType link_type, bool test_connection = false)
	{
		NySpeed 			speed	= new NySpeed;
		NyConnectionType	type	= new NyConnectionType;
		NyNodeDetails		details	= new NyNodeDetails;
		InternetAddress		myip	= getSystemIPAddress(config.fileSharingPort);
		
		speed.speed 			= config.speed;
		type.port0	 			= config.isPort0;
		type.linkType			= link_type;
		type.badPort0Connection	= test_connection;
		
		details.address 	= myip.addr;
		details.port 		= myip.port;
		
		details.clusteringKey(config.clusteringKey);
		
		conn.sendCommand(speed);
		conn.sendCommand(type);
		conn.sendCommand(details);
	}
	
	// ノード情報受信
	static void receiveNodeDetails(NyConnection conn, 
							out NySpeed speed, out NyConnectionType type,
							out NyNodeDetails details)
	{
		bool processing = true;
		
	_RECEIVE_LOOP:
		do {
			NyRawCommand command =conn.receiveCommand();
			switch (command.commandCode) {
			case NyCommandCode.SPEED:
				speed	= new NySpeed(command);
				break;
				
			case NyCommandCode.CONNECTION_TYPE:
				type = new NyConnectionType(command);
				break;
				
			case NyCommandCode.NODE_DETAILS:
				details = new NyNodeDetails(command);
				processing = false;
				break;
				
			default:
				throw new ReceiveError; // テスト接続などもこれ
			}
			//delete command;
		} while (processing);	// BBSノードの場合は多い
		
		if (speed is null 
			|| type is null
			|| details is null)
		{
			throw new ReceiveError;
		}
	}
	
	void lock()
	{
		m_lock = true;
	}
	
	void unlock()
	{
		m_lock = false;
	}
	
	// getter
	
	bool isLock()
	{
		return m_lock;
	}
	
	NyConnection connection()
	{
		return m_connection;
	}
	
	NodeDetails nodeDetails()
	{
		return m_node;
	}
	
	time_t connectTime()
	{
		return m_time;
	}
	
	time_t ctl()
	{
		return m_connection.ctl;
	}
	
	// 連想配列用のキー. ノードテーブルと同じキーという前提
	// 同じノードへは同じ種類の接続を作らないように気をつけること
	// -> 同じノードへ２つの転送リンクは張らない
	ulong aaKey()
	in {
		assert(m_node !is null);
	}
	body {
		return m_node.aaKey();
	}
	
	
	// 拡散クエリ制御タイマ
	
	bool isDiffusionRequestTime(time_t cur = 0)
	{
		if (cur == 0) {
			cur = time(null);
		}
		if (cur - m_last_diffusion_request_time > 30) {
			return true;
		}
		
		return false;
	}
	
	bool isDiffusionResponseTime(time_t cur = 0)
	{
		if (cur == 0) {
			cur = time(null);
		}
		if (cur - m_last_diffusion_response_time > 20) { // 10秒ほど余裕をおく
			return true;
		}
		
		return false;
	}
	
	void updateLastDiffusionResponseTime(time_t cur = 0)
	{
		if (cur == 0) {
			cur = time(null);
		}
		m_last_diffusion_response_time = cur;
	}
	
	void updateLastDiffusionRequestTime(time_t cur = 0)
	{
		if (cur == 0) {
			cur = time(null);
		}
		m_last_diffusion_request_time = cur;
	}
}

// コネクションテーブル
class ConnectionTable: Locker!(ulong, Connection)
{
public:
	this(SemaphoreResource resource, semaphore id)
	{
		super(resource, id);
	}
	
	override void set(ulong key, Connection conn)
	{
		signal();
		
		try {
			if (exists(key)) {
				Connection oldconn = get(key);
				if (oldconn is conn) {
					return;
				}
				// 同じコネクションは古いほうを切断しとく
				remove(key);
			}
			super.set(key, conn);
		} finally {
			release();
		}
	}
	
	override void remove(ulong key)
	{
		signal();
		try {
			if (exists(key)) {
				try {
					get(key).close();
				} catch (SocketException) {
					// ここは例外にしない
				}
				super.remove(key);
			} else {
				assert(0);
			}
		} finally {
			release();
		}
	}
	
	override bool exists(ulong key, Lock lock = Lock.BLOCK)
	{
		return super.exists(key, lock); // なんかエラー出るので入れた
	}
	
	bool exists(uint address, uint port)
	{
		signal();
		try {
			foreach (Connection e; values) {
				if ((e.nodeDetails.inetAddress.addr == address
					|| e.nodeDetails.reportAddress.addr == address)
					&& e.nodeDetails.reportAddress.port == port)
				{
					return true; // ok
				}
			}
		} finally {
			release();
		}
		
		return false;
	}
}



