﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.data.sqlite3;

// sqlite3をDで使えるようにしたもの

// 今のところ個人的に使うものだけしか入れていない


private import std.stdarg;
private import std.string;
private import std.c.string;

pragma(lib, "sqlite3.lib"); // この名前は俺が決めたもの

class SQLiteException: Exception
{
protected:
	int m_errcode;
public:
	this(int errcode, char []msg)
	{
		m_errcode = errcode;
		super(msg);
	}
	
	int errcode()
	{
		return m_errcode;
	}
}

class SQLiteBusyException: SQLiteException
{
	this(int errcode, char[] msg)
	{
		super(errcode, msg);
	}
}

// SQLiteExceptionを継承しているけど、こっちの内部的なエラー
class SQLiteInternalException: SQLiteException
{
	this(char[] msg)
	{
		super(SQLITE_OK, msg);
	}
}

class SQLiteStatement
{
protected:
	SQLite			m_parent;
	sqlite3_stmt 	*m_stmt;
	bool			m_row;
	
public:
	this(sqlite3_stmt *stmt, SQLite parent)
	{
		m_stmt 		= stmt;
		m_parent	= parent;
	}
	
	~this()
	{
		if (m_stmt != null) {
			sqlite3_finalize(m_stmt);
			m_stmt = null;
		}
	}
	
	// 値のバインド 型情報でAPIを切り替える
	void bind(int n, ...)
	{
		int result;
		for (int i = 0; i < _arguments.length; ++i) {
			/* SQLite型	*/
			// 省略
			
			/* 64bit整数	*/
			if (_arguments[i] is typeid(long)) {
				result = sqlite3_bind_int64(m_stmt, n + i, va_arg!(long)(_argptr));
			} else if (_arguments[i] is typeid(ulong)) {
				result = sqlite3_bind_int64(m_stmt, n + i, cast(long)va_arg!(ulong)(_argptr));
			} else
			/* 32bitまでの整数	*/
			if (_arguments[i] is typeid(int)) {
				result = sqlite3_bind_int(m_stmt, n + i, va_arg!(int)(_argptr));
			} else if (_arguments[i] is typeid(uint)) {
				result = sqlite3_bind_int(m_stmt, n + i, cast(int)va_arg!(uint)(_argptr));
			} else if (_arguments[i] is typeid(short)) {
				result = sqlite3_bind_int(m_stmt, n + i, va_arg!(short)(_argptr));
			} else if ( _arguments[i] is typeid(ushort)) {
				result = sqlite3_bind_int(m_stmt, n + i, cast(int)va_arg!(ushort)(_argptr));
			} else if (_arguments[i] is typeid(byte)) {
				result = sqlite3_bind_int(m_stmt, n + i, va_arg!(byte)(_argptr));
			} else if (_arguments[i] is typeid(ubyte)) {
				result = sqlite3_bind_int(m_stmt, n + i, cast(int)va_arg!(ubyte)(_argptr));
			} else 
			/* 不動小数点数	*/
			if (_arguments[i] is typeid(real)) {
				result = sqlite3_bind_double(m_stmt, n + i, cast(double)va_arg!(real)(_argptr));
			} else if (_arguments[i] is typeid(double)) {
				result = sqlite3_bind_double(m_stmt, n + i, va_arg!(double)(_argptr));
			} else if (_arguments[i] is typeid(float)) {
				result = sqlite3_bind_double(m_stmt, n + i, va_arg!(float)(_argptr));
			} else
			/* utf-8テキスト	*/
			if (_arguments[i] is typeid(char[])) {
				char[] text = va_arg!(char[])(_argptr);
				result = sqlite3_bind_text(m_stmt, n + i, text.ptr, text.length, SQLITE_TRANSIENT);
			}
			// utf-16テキスト	*/
			else if (_arguments[i] is typeid(wchar[])) {
				wchar[] text16 = va_arg!(wchar[])(_argptr);
				result = sqlite3_bind_text16(m_stmt, n + i, text16.ptr, text16.length, SQLITE_TRANSIENT);
			} else if (_arguments[i] is typeid(ubyte[])) {
			// Blob
				ubyte[] blob = va_arg!(ubyte[])(_argptr);
				result = sqlite3_bind_blob(m_stmt, n + i, blob.ptr, blob.length, SQLITE_TRANSIENT);
			}
			else if (_arguments[i] is typeid(void *)) {
				void *p = va_arg!(void *)(_argptr);
				if (p == null) {
					result = sqlite3_bind_null(m_stmt, n + i);
				} else {
					assert(0);
				}
			} else {
				assert(0);
			}
			
			if (result != SQLITE_OK) {
				m_parent.throwCurrentException();
			}
		}
	}
	
	void step()
	{
		int result = sqlite3_step(m_stmt);
		if (result == SQLITE_ROW) {
			m_row = true;
		} else {
			m_row = false;
		}
		if (result != SQLITE_DONE && result != SQLITE_ROW) {
			m_parent.throwCurrentException();
		}
	}
	
	bool hasRow()
	{
		return m_row;
	}
	
	void reset()
	{
		int result = sqlite3_reset(m_stmt);
		if (result != SQLITE_OK) {
			m_parent.throwCurrentException();
		}
	}
	
	int dataCount()
	{
		int result = sqlite3_data_count(m_stmt);
		
		return result;
	}
	
	int columnCount()
	{
		int result = sqlite3_column_count(m_stmt);
		
		return result;
	}
	
	// 値の取り出し テンプレート
	template fetch(T) {
		T fetch(int col)
		{
			static if (is(T == long) || is(T == ulong)) {
				// 64bit整数型
				return cast(T)sqlite3_column_int64(m_stmt, col);
			} else static if (is(T == real) || is(T == double) || is(T == float)) {
				// 浮動小数点型
				return cast(T)sqlite3_column_double(m_stmt, col);
			} else static if (is(T: int)) {
				// その他intへキャスト可能な整数型
				return cast(T)sqlite3_column_int(m_stmt, col);
			} else static if (is(T == char[])) {
				// utf8 テキスト
				return (cast(char *)sqlite3_column_text(m_stmt, col))[0 .. sqlite3_column_bytes(m_stmt, col)];
			} else static if (is(T == wchar[])) {
				// utf16 テキスト
				return (cast(wchar*)sqlite3_column_text16(m_stmt, col))[0 .. sqlite3_column_bytes16(m_stmt, col)];
			} else static if (is(T == ubyte[])) {
				// blob
				return cast(T)sqlite3_column_blob(m_stmt, col)[0 .. sqlite3_column_bytes(m_stmt, col)];
			} else {
				
				if (sqlite3_column_bytes(m_stmt, col) < T.sizeof) {
					throw new SQLiteInternalException("SQLiteInternalException: データサイズ小さすぎ");
				}
				
				return *cast(T *)sqlite3_column_blob(m_stmt, col);
			}
		}
	}

	// なぜかsqlite3_clear_bindingsがリンクエラー: 未実装かなくなったっぽい
/*	void clear()
	{
		int result = sqlite3_clear_bindings(m_stmt);
		if (result != SQLITE_OK) {
			m_parent.throwCurrentException();
		}
	}
*/	
	void finalize()
	{
		int result = sqlite3_finalize(m_stmt);
		
		if (result != SQLITE_OK) {
			m_parent.throwCurrentException();
		}
		m_stmt = null;
	}
	
}


class SQLite 
{
protected:
	sqlite3		*m_conn;
	
public:
	void throwCurrentException()
	{
		int errcode		= sqlite3_errcode(m_conn);
		char 	*cmsg 	= sqlite3_errmsg(m_conn);
		char[]	dmsg 	= cmsg[0 .. strlen(cmsg)];
		
		switch (errcode) {
		case SQLITE_BUSY:
			throw new SQLiteBusyException(errcode, dmsg);
		default:
			throw new SQLiteException(errcode, dmsg);
		}
	}
	
	this(char[] filename,
		int cache_size = 42000,
		char[] synchronous = "OFF",
		char[] temp_store = "MEMORY")
	{
		open(filename, cache_size, synchronous, temp_store);
	}
	
	~this()
	{
		if (m_conn != null) {
			sqlite3_close(m_conn);
			m_conn = null;
		}
	}
	
	void open(char[] filename, int cache_size, char[] synchronous, char[] temp_store)
	in {
		assert(m_conn == null);
	}
	body {
		char[] param = std.string.format(
			"PRAGMA cache_size = %d;"
			"PRAGMA synchronous = %s;"
			"PRAGMA temp_store = %s;",
			cache_size, synchronous, temp_store
		);
		int result = sqlite3_open(toStringz(filename), &m_conn);
		if (result != SQLITE_OK) {
			throwCurrentException();
		}
		exec(param);
	}
	
	void close()
	in {
		assert(m_conn != null);
	}
	body {
		int result = sqlite3_close(m_conn);
		if (result != SQLITE_OK) {
			throwCurrentException();
		}
		m_conn = null;
	}
	
	/* トランザクション	*/
	
	void begin()
	{
		exec("BEGIN");
	}
	
	void end()
	{
		exec("END");
	}
	
	void rollback()
	{
		exec("ROLLBACK");
	}
	
	void commit()
	{
		exec("COMMIT");
	}
	
	void exec(char[] sql)
	{
		int result = sqlite3_exec(m_conn, toStringz(sql), null, null, null);
		
		if (result != SQLITE_OK) {
			throwCurrentException();
		}
	}
	
	// 実行: 効率を考えた結果、callbackはラップせず直接通すようにした
	void exec_ex(char[] sql, sqlite3_callback callback, void *argv)
	{
		int result = sqlite3_exec(m_conn, toStringz(sql), callback, argv, null);
		
		if (result != SQLITE_OK) {
			throwCurrentException();
		}
	}
	
	// 準備
	SQLiteStatement prepare(char[] sql)
	{
		sqlite3_stmt *stmt;
		int result = sqlite3_prepare(m_conn, sql.ptr, sql.length, &stmt, cast(char**)null);
		if (result != SQLITE_OK) {
			throwCurrentException();
		}
		
		return new SQLiteStatement(stmt, this);
	}
	
	
}

import std.stdio;
char[] escapeSQLText(char[] text)
{
	char[] esc = std.string.replace(text, "'", "''");
	esc = std.string.replace(esc, "%", "\\%");
	esc = std.string.replace(esc, "_", "\\_");
	
	return esc;
}
/*
private char[] wreplace(char[] text, char[] from, char[] to)
{
char[] replace(char[] s, char[] from, char[] to)
	{
	wchar[] p;
	int i;
	int istart;

	//printf("replace('%.*s','%.*s','%.*s')\n", s, from, to);
	if (from.length == 0)
	return s;
	istart = 0;
	while (istart < s.length)
	{
		i = find(s[istart .. s.length], from);
	if (i == -1)
	{
	    p ~= s[istart .. s.length];
	    break;
	}
	p ~= s[istart .. istart + i];
	p ~= to;
	istart += i + from.length;
	}
	return p;
}
*/

/* 以下SQLite3 のC/C++ API	*/


extern(C) {
	// strong typedef 
	struct sqlite3		{ /* abstract */ };
	struct sqlite3_stmt	{ /* abstract */ };
		
	const int SQLITE_OK			= 0;    /* Successful result */
	const int SQLITE_ERROR		= 1;    /* SQL error or missing database */
	const int SQLITE_INTERNAL	= 2;    /* An internal logic error in SQLite */
	const int SQLITE_PERM		= 3;    /* Access permission denied */
	const int SQLITE_ABORT		= 4;    /* Callback routine requested an abort */
	const int SQLITE_BUSY		= 5;    /* The database file is locked */
	const int SQLITE_LOCKED		= 6;    /* A table in the database is locked */
	const int SQLITE_NOMEM		= 7;    /* A malloc() failed */
	const int SQLITE_READONLY	= 8;    /* Attempt to write a readonly database */
	const int SQLITE_INTERRUPT	= 9;    /* Operation terminated by sqlite_interrupt() */
	const int SQLITE_IOERR		= 10;   /* Some kind of disk I/O error occurred */
	const int SQLITE_CORRUPT	= 11;   /* The database disk image is malformed */
	const int SQLITE_NOTFOUND	= 12;   /* (Internal Only) Table or record not found */
	const int SQLITE_FULL		= 13;   /* Insertion failed because database is full */
	const int SQLITE_CANTOPEN	= 14;   /* Unable to open the database file */
	const int SQLITE_PROTOCOL	= 15;   /* Database lock protocol error */
	const int SQLITE_EMPTY		= 16;   /* (Internal Only) Database table is empty */
	const int SQLITE_SCHEMA		= 17;   /* The database schema changed */
	const int SQLITE_TOOBIG		= 18;   /* Too much data for one row of a table */
	const int SQLITE_CONSTRAINT	= 19;   /* Abort due to constraint violation */
	const int SQLITE_MISMATCH	= 20;   /* Data type mismatch */
	const int SQLITE_MISUSE		= 21;   /* Library used incorrectly */
	const int SQLITE_NOLFS		= 22;   /* Uses OS features not supported on host */
	const int SQLITE_AUTH		= 23;   /* Authorization denied */
	const int SQLITE_ROW		= 100;  /* sqlite_step() has another row ready */
	const int SQLITE_DONE		= 101;  /* sqlite_step() has finished executing */
	int sqlite3_open(
		char *filename,		/* Database filename (UTF-8) */
		sqlite3 **ppDb		/* OUT: SQLite db handle */
	);
	
	alias int function(void *, int, char **, char **)	sqlite3_callback;
	
	/* 実行	*/
	int sqlite3_exec(
		sqlite3 *,				/* An open database */
		char *sql,				/* SQL to be executed */
		sqlite3_callback,		/* Callback function */
		void *,					/* 1st argument to callback function */
		char **errmsg			/* Error msg written here */
	);
	
	char *sqlite3_errmsg(sqlite3 *);
	int sqlite3_errcode(sqlite3 *);
	
	/* 準備	*/
	int sqlite3_prepare(
		sqlite3 *db,			/* Database handle */
		char *zSql,				/* SQL statement, UTF-8 encoded */
		int nBytes,				/* Length of zSql in bytes. */
		sqlite3_stmt **ppStmt,	/* OUT: Statement handle */
		char **pzTail     		/* OUT: Pointer to unused portion of zSql */
	);
	
	/* prepare後にデータをセットしたりするもの */
	
	alias void function(void *)		sqlite3_disposer;
	const sqlite3_disposer SQLITE_STATIC	= cast(sqlite3_disposer)0;
	const sqlite3_disposer SQLITE_TRANSIENT	= cast(sqlite3_disposer)-1;
	
	int sqlite3_bind_blob(sqlite3_stmt *, int, void *, int n, sqlite3_disposer disporser);
	int sqlite3_bind_double(sqlite3_stmt*, int, double);
	int sqlite3_bind_int(sqlite3_stmt*, int, int);
	int sqlite3_bind_int64(sqlite3_stmt*, int, long);
	int sqlite3_bind_null(sqlite3_stmt*, int);
	int sqlite3_bind_text(sqlite3_stmt*, int, char *, int n, sqlite3_disposer disporser);
	int sqlite3_bind_text16(sqlite3_stmt*, int, void *, int n, sqlite3_disposer disporser);
	
	/* パラメータの位置取得	*/
	char *sqlite3_bind_parameter_name(sqlite3_stmt*, int n);
	int sqlite3_bind_parameter_index(sqlite3_stmt*, char *zName);
	
	/* 	ステートメントをリセット.
		これは、prepareを呼んだ後の実行準備が完了した状態に戻す
	*/
	int sqlite3_reset(sqlite3_stmt *);
	
	/* ステートメントをクリア	*/
	int sqlite3_clear_bindings(sqlite3_stmt*);
	
	/* SQLを１つ実行 */
	int sqlite3_step(sqlite3_stmt *);
	
	/*	変更されたROWS数.
	 	DELETEは、WHEREをつけないとカウントされない.
		DELETE FROM TABLE;の場合、WHERE 1をつけるとカウントする
	*/
	int sqlite3_changes(sqlite3*);
	
	/* 返されたROW数	*/
	int sqlite3_data_count(sqlite3_stmt *);
	
	
	/* 実行結果を取得すもの	*/
	void *sqlite3_column_blob(sqlite3_stmt*, int iCol);
	int sqlite3_column_bytes(sqlite3_stmt*, int iCol); // blobをutf8へ変換したときのサイズ
	int sqlite3_column_bytes16(sqlite3_stmt*, int iCol);// blobをutf16へ変換したときのサイズ
	double sqlite3_column_double(sqlite3_stmt*, int iCol);
	int sqlite3_column_int(sqlite3_stmt*, int iCol);
	long sqlite3_column_int64(sqlite3_stmt*, int iCol);
	char *sqlite3_column_text(sqlite3_stmt*, int iCol);
	void *sqlite3_column_text16(sqlite3_stmt*, int iCol);
	
	/* カラム数	*/
	int sqlite3_column_count(sqlite3_stmt *);
	
	const int SQLITE_INTEGER	= 1;
	const int SQLITE_FLOAT		= 2;
	const int SQLITE_TEXT		= 3;
	const int SQLITE_BLOB		= 4;
	const int SQLITE_NULL		= 5;
	int sqlite3_column_type(sqlite3_stmt*, int iCol);
	
	/* 	ステートメントをデリート*
		SQLを実行していないか全て成功している場合は、
		SQLITE_OKを返す。そうでない場合エラーコードが返る.
		各ステートメント実行後は必ずこれを実行しないとクローズに失敗することになる.
	*/
	int sqlite3_finalize(sqlite3_stmt *);
	
	/* クローズ	*/
	int sqlite3_close(sqlite3 *);
	
	/* テーブル内容取得 */
	int sqlite3_get_table(
		sqlite3*,				/* An open database */
		char *sql,				/* SQL to be executed */
		char ***resultp,		/* Result written to a char *[]  that this points to */
		int *nrow,				/* Number of result rows written here */
		int *ncolumn,			/* Number of result columns written here */
		char **errmsg			/* Error msg written here */
	);
	void sqlite3_free_table(char **result);
}

debug(pes_data_sqlite3)
{
	int main()
	{
		return 0;
	}
}
