/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.net.winny.command;

// Winny Network Protocol impl
// by C.xxxxxxxx

private import std.string;
private import std.socket;
private import std.c.time;
private import pes.rc4;
private import pes.byteorder;
private import pes.prompt;
private import pes.dump;
private import pes.checksum;
private import pes.queue;
private import pes.net.winny.connection;
private import pes.net.winny.key;

private import std.stdio; // デバッグ用あとでけす

enum NyCommandCode: ubyte
{
	PROTOCOL_HEADER 				= 0,
	SPEED							= 1,
	CONNECTION_TYPE					= 2,
	NODE_DETAILS					= 3,
	ANOTHER_NODE					= 4,
	BBS_PORT						= 5,
	DIFFUSION_REQUEST				= 10,
	FILE_REQUEST					= 11,
	CONDITIONAL_DIFFUSION_REQUEST	= 12,
	QUERY							= 13,
	FILE_RESPONSE					= 21,
	CLOSE							= 31,
	CONNECTED_LIMITATION			= 32,
	WRONG_LISTENING_PORT			= 33,
	REJECT							= 34,
	SLOW_RATE						= 35,
	LIAR							= 36,
	LOW_VERSION						= 97,
	
	//
}
alias			uint CommandLength;

// exception
class NyBadCommand: SocketException // なぜかソケット
{
	this(char[] msg)
	{
		super(msg);
	}
}

ubyte[] toStep2Key(ubyte[] step1_key)
{
	ubyte[] step2_key = step1_key.dup;
	
	for (int i = 1; i < step2_key.length; ++i) {
		step2_key[i] ^= 0x39;
//		step2_key[i] ^= step2_key[i-1]; poeny test
	}
	
	return step2_key;
}

char[] getCString(ubyte[] data)
{
	char[] s;
	
	foreach (ubyte b; data) {
		if (b == 0) {
			break;
		}
		s ~= cast(char)b;
	}
	
	return s;
}

static ubyte[] uIntToUByteArray(CommandLength command_length)
{
	ubyte[] byte_array = new ubyte[CommandLength.sizeof];// 4
	
	// リトルエンディアン
	byte_array[3] = (command_length & 0xff000000) >> 24;
	byte_array[2] = (command_length & 0x00ff0000) >> 16;
	byte_array[1] = (command_length & 0x0000ff00) >> 8;
	byte_array[0] = (command_length & 0x000000ff);
	
	return byte_array;
}

// Winnyコマンド
abstract class NyCommand
{
protected:
	CommandLength	m_length;
	NyCommandCode	m_command_code;
	ubyte[]			m_data;
	ubyte[]			m_recv_buffer;
	
public:
	enum UnpackHow: int {
		HEADER_ONLY = 0,
		ALL			= 1
	};
	
	abstract void unpack(NyRawCommand raw_command);
	
	this() { m_recv_buffer = null;}
	
	void unpack(ubyte[] command_block)
	{
		unpack(command_block, UnpackHow.ALL);
	}
	
	ubyte[] unpack(ubyte[] command_block, UnpackHow how)
	{
		m_recv_buffer = command_block;
		
		if (command_block.length < headerLength) {
			throw new NyBadCommand("NyCommand: command is too small");
		}
		m_length = *cast(CommandLength *)command_block.ptr;
		
		if (m_length == 0) {
			throw new NyBadCommand("NyCommand: command code is nothing");
		}
		m_command_code 		= cast(NyCommandCode)command_block[CommandLength.sizeof];
		
		if (command_block.length < headerLength + m_length -1) {
			throw new NyBadCommand("NyCommand: length is too small");
		}
		if (how == UnpackHow.HEADER_ONLY) {
			m_data = null;
			if (m_length == 1) {
				return null;
			}
			
			return command_block[headerLength .. headerLength + m_length-1];
		}
		
		if (m_length > uint.sizeof + NyCommandCode.sizeof) {
			m_data = command_block[headerLength .. headerLength + m_length-1];
		} else {
			m_data = null;
		}
		
		return m_data;
	}
	
	void assign(NyCommand command)
	{
		m_length 		= command.length;
		m_command_code 	= command.commandCode;
		m_data			= command.m_data; //.dup
	}
	
	ubyte[] pack()
	{
		ubyte[] packet = new ubyte[CommandLength.sizeof + NyCommandCode.sizeof + m_data.length];
		
		uint pos;
		
		// length
		packet[0 .. CommandLength.sizeof] =  uIntToUByteArray(NyCommandCode.sizeof + m_data.length);
		pos += CommandLength.sizeof;
		
		// command code
		packet[pos .. pos + NyCommandCode.sizeof] = m_command_code;
		pos += NyCommandCode.sizeof;
		
		// data
		packet[pos .. pos + m_data.length] = m_data;
		
		if (m_data.length != 0) {
			delete m_data;
			m_data = null;
		}
		
		return packet;
	}
	
	uint length()
	{
		return m_length;
	}
	
	
	static uint headerLength()
	{
		return CommandLength.sizeof + NyCommandCode.sizeof;
	}
	
	
	uint dataLength()
	{
		return m_data.length;
	}
	
	
	NyCommandCode commandCode()
	{
		return m_command_code;
	}
	
	
	void commandCode(NyCommandCode command_code)
	{
		m_command_code = command_code;
	}
	
	
	int nextIndex()
	{
		return CommandLength.sizeof + NyCommandCode.sizeof + length() -1;
	}
	
	
	char[] toString()
	{
		return format("CommandCode(%d) Length(%d)\n", cast(ubyte)commandCode, length);
	}
	
	void destroy()
	{
	}
	
	void clearBuffer()
	{
		if (m_recv_buffer.length != 0) {
			delete m_recv_buffer;
		}
		m_recv_buffer 	= null;
		m_data			= null;
	}

}

class NyRawCommand: NyCommand
{
public:
	this(ubyte[] command_block)
	{
		unpack(command_block);
	}
	
	this() {}
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
	}
	
	override void unpack(ubyte[] command_block)
	{
		ubyte[] data = super.unpack(command_block, UnpackHow.HEADER_ONLY);
		m_data = data;// .dup;
	}
	
	ubyte[] data()
	{
		return m_data; // do not dup
	}
}

// Command00
// Winnyプロトコルヘッダ
class NyProtocolHeader: NyCommand
{
protected:
	// 認証データ用RC4暗号キー
	static const ubyte[] CERT_KEY 	= [ 0x39, 0x38, 0x37, 0x38, 0x39, 0x61, 0x73, 0x6a];
	// バージョン
	uint		m_minor;
	// バージョン
	char[]		m_major;
	
	char[] minorToString()
	{
		return std.string.toString(m_minor);
	}
	
	ubyte[] certCrypt(ubyte[] data)
	{
		return pes.rc4.RC4.crypt(CERT_KEY, data);
	}
	
	void unpackInternal(ubyte[] data)
	{
		ubyte[] cert;
		
		if (length < m_minor.sizeof) {
			throw new NyBadCommand("Command is too small");
		}
		
		// 復号
		cert = certCrypt(data);
		
		// マイナーバージョン (32bit little endian)
		m_minor =  *(cast(uint *)&cert[0]);
		
		// バージョン文字列
		m_major	= mbsToUTF8(cert[m_minor.sizeof .. super.length -1]);
		
//		delete data;
		m_data = null;
		clearBuffer();
	}
	
public:
//	static const char[] WINNY_MAJOR = "Winny Ver2.0b1 ";
	static const char[]	WINNY_MAJOR	= "Winny Ver2.0b1 (poeny)";
//	static const char[]	WINNY_MAJOR	= "poeny Ver0";
	
	static const uint	WINNY_MINOR	= 12710;
//	static const uint	WINNY_MINOR	= 0;
	
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}

	/* 
	 * getter 
	 *
	 */
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal(raw_command.data);
	}
	
	override void unpack(ubyte[] command_block)
	{
		ubyte[] data;
		
		data = super.unpack(command_block, UnpackHow.HEADER_ONLY);
		
		unpackInternal(data);
//		delete data;
	}
	
	uint minor()
	{
		return m_minor;
	}
	
	char[] major()
	{
		return m_major;
	}
	
	char[] toString()
	{
		return major ~ "(" ~ minorToString ~ ")";
	}
	
	/*
	 * setter 
	 *
	 */
	
	override ubyte[] pack()
	{
		m_data		=  uIntToUByteArray(m_minor);
		m_data 		~= cast(ubyte[])m_major;
		
		// 暗号化
		m_data		= certCrypt(m_data);
		
		commandCode	=  NyCommandCode.PROTOCOL_HEADER;
		
		return super.pack();
	}
	
	void minor(uint minor_version) 
	{
		m_minor = minor_version;
	}
	
	void major(char[] major_version)
	{
		m_major = major_version.dup;
	}
}

// Command01
// 回線速度の通知
class NySpeed: NyCommand
{
protected:
	float	m_speed;
	
	void unpackInternal(ubyte[] data)
	{
		
		if (data.length < m_speed.sizeof) {
			throw new NyBadCommand("Speed: length is too small");
		}
		
		m_speed = *cast(float *)data;
		
//		delete data;
		m_data = null;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	// getter
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal(raw_command.data);
	}
	
	override void unpack(ubyte[] command_block)
	{
		ubyte[] data = super.unpack(command_block, UnpackHow.HEADER_ONLY);
		
		unpackInternal(data);
	}
	
	
	uint speed()
	{
		return cast(int)m_speed;
	}
	
	// setter
	
	ubyte[] pack()
	{
		m_data = new ubyte[float.sizeof];
		*(cast(float *)(m_data.ptr)) = speed;
		
		commandCode	=  NyCommandCode.SPEED;
		
		return super.pack();
	}
	
	void speed(uint line_speed)
	{
		m_speed = cast(float)cast(int)line_speed;
	}
}


// Command02
// コネクション種別の通知
class NyConnectionType: NyCommand
{
public:
	enum LinkType: ubyte 
	{
		SEARCH 		= 0,
		TRANSFER,
		BBS_SEARCH
	}
	
protected:
	LinkType 	m_linktype;
	bool		m_is_port0;
	bool		m_is_badport0;
	bool		m_is_bbslink;
	
	void unpackInternal(ubyte[] data)
	{
		if (data.length < 
				m_linktype.sizeof 
			+ 	m_is_port0.sizeof 
			+ 	m_is_badport0.sizeof 
			+ 	m_is_bbslink.sizeof)
		{
			throw new NyBadCommand("ConnectionType: length is too small");
		}
		switch (data[0]) {
		case LinkType.SEARCH:
		case LinkType.TRANSFER:
		case LinkType.BBS_SEARCH:
			m_linktype = cast(LinkType)data[0];
			break;
		default:
			throw new NyBadCommand(format("ConnectionType: unknown LinkType `%d'", cast(int)data[0]));
		}
		m_is_port0 		= data[1] ? true: false;
		m_is_badport0 	= data[2] ? true: false;
		m_is_bbslink 	= data[3] ? true: false;
		
//		delete data;
		m_data = null;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	// getter
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal(raw_command.data);
	}
	
	override void unpack(ubyte[] command_block)
	{
		ubyte[] data = super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal(data);
	}
	
	bool isSearchLink()
	{
		return m_linktype == LinkType.SEARCH;
	}
	
	bool isTransferLink()
	{
		return m_linktype == LinkType.TRANSFER;
	}
	
	bool isBBSSearchLink()
	{
		return m_linktype == LinkType.BBS_SEARCH;
	}
	
	LinkType linkType()
	{
		return m_linktype;
	}
	
	bool isPort0()
	{
		return m_is_port0;
	}
	
	bool isBadPort0Connection()
	{
		return m_is_badport0;
	}
	
	bool isNormalConnection()
	{
		return !isBadPort0Connection;
	}
	
	bool isBBSLink()
	{
		return m_is_bbslink;
	}
	
	bool isFileSharingLink()
	{
		return !isBBSLink;
	}
	
	// settter
	
	override ubyte[] pack()
	{
		uint data_size 		=	m_linktype.sizeof 
							+ 	m_is_port0.sizeof 
							+ 	m_is_badport0.sizeof 
							+ 	m_is_bbslink.sizeof;
		m_data = new ubyte[data_size];
		
		m_data[0] = cast(ubyte)m_linktype;
		m_data[1] = m_is_port0 ? 1: 0;
		m_data[2] = m_is_badport0 ? 1: 0;
		m_data[3] = m_is_bbslink ? 1: 0;
		
		commandCode = NyCommandCode.CONNECTION_TYPE;
		
		return super.pack();
	}
	
	void linkType(LinkType link_type)
	{
		m_linktype = link_type;
	}
	
	void port0(bool flag)
	{
		m_is_port0 = flag;
	}
	
	void badPort0Connection(bool flag)
	{
		m_is_badport0 = flag;
	}
	
	void bbsLink(bool flag)
	{
		m_is_bbslink = flag;
	}
}

// Command03
// 自ノード情報の通知
class NyNodeDetails: NyCommand
{
protected:
	uint			m_address;
	uint			m_listener_port;
	char[]			m_ddns_name;
	char[][3]		m_clustering_key;
	
	struct PackStruct {
	align(1):
		uint		ip_address;
		uint		listener_port;
		ubyte		ddns_length;
		ubyte[3]	cluster_length;
	}
	
	void unpackInternal(ubyte[] data)
	{
		uint 			ddns_len;
		uint			offset = 0;
		uint[3]			cluster_len;
		PackStruct		*pack;
		
		if (data.length < PackStruct.sizeof) {
			throw new NyBadCommand("NodeDetails: length is too small");
		}
		pack = cast(PackStruct *)data.ptr;
		
		if (data.length !=
				PackStruct.sizeof 
			+ 	pack.ddns_length 
			+ 	pack.cluster_length[0]
			+ 	pack.cluster_length[1]
			+ 	pack.cluster_length[2])
		{
			throw new NyBadCommand("NodeDetails: illegal length");
		}
		
		m_address		= networkToHost32(pack.ip_address);
		m_listener_port	= pack.listener_port;
		m_ddns_name		= cast(char[])data[PackStruct.sizeof 
											.. PackStruct.sizeof + pack.ddns_length].dup;
		
		offset = PackStruct.sizeof + pack.ddns_length;
		for (int i = 0; i < m_clustering_key.length; ++i) {
			m_clustering_key[i] 	=  mbsToUTF8(data[offset .. offset + pack.cluster_length[i]]).dup;
			offset 				+= pack.cluster_length[i];
		}
		
//		delete m_data;
		m_data = null;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	// getter
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal(raw_command.data);
	}
	
	override void unpack(ubyte[] command_block)
	{
		ubyte[] data = super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal(data);
	}
	
	uint address()
	{
		return m_address;
	}
	
	uint port()
	{
		return m_listener_port;
	}
	
	char[] ddns()
	{
		return m_ddns_name;
	}
	
	char[] clusterWord(int index)
	{
		return m_clustering_key[index];
	}
	
	char[][] clusteringKey()
	{
		return m_clustering_key;
	}
	
	// setter
	
	override ubyte[] pack()
	{
		PackStruct 	pack;
		ubyte[] 	ddns_name, mbs_word;
		ubyte[][3] 	clustering_word;
		
		ddns_name = cast(ubyte[])m_ddns_name.dup;
		for (int i = 0; i < m_clustering_key.length; ++i) {
			if (m_clustering_key !is null) {
				clustering_word[i]	= utf8ToMBS(m_clustering_key[i]);
			}
		}
		
		pack.ip_address 		=  networkToHost32(m_address);
		pack.listener_port	 	=  m_listener_port;
		pack.ddns_length		=  ddns_name.length;
		pack.cluster_length[0]	=  clustering_word[0].length;
		pack.cluster_length[1]	=  clustering_word[1].length;
		pack.cluster_length[2]	=  clustering_word[2].length;
		
		m_data =  (cast(ubyte *)(&pack))[0 .. PackStruct.sizeof].dup;
		m_data ~=  cast(ubyte[])ddns_name;
		m_data ~= clustering_word[0].dup;
		m_data ~= clustering_word[1].dup;
		m_data ~= clustering_word[2].dup;
		
		commandCode = NyCommandCode.NODE_DETAILS;
		
		return super.pack();
	}
	
	void address(uint ip_address)
	{
		m_address = ip_address;
	}
	
	void port(ushort tcp_port)
	{
		m_listener_port = tcp_port;
	}
	
	void ddns(char[] ddns_name)
	{
		m_ddns_name = ddns_name.dup;
	}
	
	// index -> 0 .. 2
	void clusteringWord(int index, char[] clustering_word)
	{
		m_clustering_key[index] = clustering_word.dup;
	}
	
	void clusteringKey(char[][] keys)
	in {
		assert(keys.length == m_clustering_key.length);
	}
	body {
		for (int i = 0; i < m_clustering_key.length; ++i) {
			m_clustering_key[i] = keys[i].dup;
		}
	}
}

// Command04
// 他ノード情報通知

class NyAnotherNode: NyCommand
{
protected:
	struct PackStruct {
	align(1):
		uint		ip_address;
		uint		file_sharing_port;
		uint		bbs_port;
		ubyte		is_bbs_node;
		uint		speed;
		ubyte[3]	clustering_word_length;
	}
	
	PackStruct 	m_pack;
	char[][3]	m_clustering_key;
	
	void unpackInternal(ubyte[] data)
	{
		int		offset;
		
		if (data.length < PackStruct.sizeof) {
			throw new NyBadCommand("NyAnotherNode: command is too small");
		}
		
		m_pack = *(cast(PackStruct *)data.ptr);
		m_pack.ip_address = hostToNetwork32(m_pack.ip_address);
		
		offset = 0;
		for (int i = 0; i < m_clustering_key.length; ++i) {
			if (m_pack.clustering_word_length[i] != 0) {
				m_clustering_key[i] = mbsToUTF8(data[PackStruct.sizeof + offset 
														.. PackStruct.sizeof + offset + m_pack.clustering_word_length[i]]);
				offset += m_pack.clustering_word_length[i];
			}
		}
		
//		delete m_data;
		m_data = null;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	// getter
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal(raw_command.data);
	}
	
	void unpack(ubyte[] command_block)
	{
		ubyte[]	data = super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal(data);
	}
	
	uint address()
	{
		return m_pack.ip_address;
	}
	
	uint fileSharingPort()
	{
		return m_pack.file_sharing_port;
	}
	
	uint bbsPort()
	{
		return m_pack.bbs_port;
	}
	
	bool isBBSNode()
	{
		return m_pack.is_bbs_node == 1 ? true: false;
	}
	
	uint speed()
	{
		return m_pack.speed;
	}
	
	char[][] clusteringKey()
	{
		return m_clustering_key;
	}
	
	// setter
	
	ubyte[] pack()
	{
		PackStruct pack = m_pack;
		
		m_data = null;
		for (int i = 0; i < m_clustering_key.length; ++i) {
			if (m_clustering_key[i].length != 0) {
				ubyte[] mbs_word = utf8ToMBS(m_clustering_key[i]);
				
				m_data ~= mbs_word;
				pack.clustering_word_length[i] = mbs_word.length;
			}
		}
		pack.ip_address = networkToHost32(pack.ip_address);
		
		m_data = (cast(ubyte *)&pack)[0 .. pack.sizeof].dup ~ m_data;
		
		commandCode = NyCommandCode.ANOTHER_NODE;
		
		return super.pack();
	}
	
	void address(uint ip_address)
	{
		m_pack.ip_address = ip_address;
	}
	
	void fileSharingPort(uint port)
	{
		m_pack.file_sharing_port = port;
	}
	
	void bbsPort(uint port)
	{
		m_pack.bbs_port = port;
	}
	
	void isBBSNode(bool is_bbs_node)
	{
		m_pack.is_bbs_node = is_bbs_node == true ? 1: 0;
	}
	
	void speed(uint line_speed)
	{
		m_pack.speed = line_speed;
	}
	
	void clusteringKey(int index, char[] key)
	in {
		assert(0 <= index && index < m_clustering_key.length);
	}
	body {
		m_clustering_key[index] = key.dup;
		m_pack.clustering_word_length[index] = key.length;
	}
	
	void clusteringKey(char[][] keys)
	in {
		assert(keys.length == m_clustering_key.length);
	}
	body {
		for (int i = 0; i < m_clustering_key.length; ++i) {
			m_clustering_key[i] = keys[i];
			m_pack.clustering_word_length[i] = keys[i].length;
		}
	}
	
	override void destroy()
	{
//		for (int i = 0; i < m_clustering_key.length; ++i) {
//			delete m_clustering_key[i];
//		}
		super.destroy();
	}
}

// Command05
// BBSポート番号通知
class NyBBSPort: NyCommand
{
protected:
	uint	m_bbs_port;
	
	void unpackInternal(ubyte[] data)
	{
		if (data.length < m_bbs_port.sizeof) {
			throw new NyBadCommand("NyBBSPort: command is too small");
		}
		m_bbs_port = *(cast(uint *)data.ptr);
		
//		delete m_data;
		m_data = null;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	// getter
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal(raw_command.data);
	}
	
	override void unpack(ubyte[] command_block)
	{
		ubyte[] data = super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal(data);
	}
	
	uint port()
	{
		return m_bbs_port;
	}
	
	// setter
	
	ubyte[] pack()
	{
		m_data = uIntToUByteArray(m_bbs_port);
		
		commandCode = NyCommandCode.BBS_PORT;
		
		return super.pack();
	}
	
	void port(uint bbs_port)
	{
		m_bbs_port = bbs_port;
	}
}

// Command10
// 拡散クエリ送信要求
class NyDiffusionRequest: NyCommand
{
protected:
	void unpackInternal()
	{
//		delete m_data;
		m_data = null;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal();
	}
	
	override void unpack(ubyte[] command_block)
	{
		super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal();
	}
	
	override ubyte[] pack()
	{
		commandCode	= NyCommandCode.DIFFUSION_REQUEST;
		m_data		= null;
		
		return super.pack();
	}
	
}

// Command12
// 条件付拡散クエリ転送要求
class NyConditionalDiffusionRequest: NyCommand
{
protected:
	struct PackStruct {
		ubyte[255]	keyword;
		ubyte[16]	trip;
		uint		queryId;
	}
	PackStruct m_packet;
	
	void unpackInternal(ubyte[] data)
	{
		if (data.length != PackStruct.sizeof) {
			throw new NyBadCommand("NyConditianlDiffusionRequest: command is too small");
		}
		m_packet = *cast(PackStruct *)data.ptr;
		
//		delete m_data;
		m_data = null;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this() {}
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal(raw_command.data);
	}
	
	void unpack(ubyte[] command_block)
	{
		ubyte[]	data = super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal(data);
	}
	
	override ubyte[] pack()
	{
		commandCode	= NyCommandCode.CONDITIONAL_DIFFUSION_REQUEST;
		m_data		= (cast(ubyte *)&m_packet)[0 .. PackStruct.sizeof].dup;
		
		return super.pack();
	}
	
	void keyword(char[] key)
	in {
		assert(key.length < m_packet.keyword.length);
	}
	body {
		if (key.length == 0) {
			m_packet.keyword[] = 0;
		} else {
			ubyte[] mbs = utf8ToMBS(key);
			m_packet.keyword[0 .. mbs.length] 			= mbs;
			m_packet.keyword[key.length-1 .. length]	= 0;
		}
	}
	
	void trip(char[] trip)
	in {
		assert(trip.length < m_packet.trip.length);
	}
	body {
		if (trip.length == 0) {
			m_packet.trip[] = 0;
		} else {
			m_packet.trip[0 .. trip.length] 			= cast(ubyte[])trip;
			m_packet.trip[trip.length - 1 .. length]	= 0;
		}
	}
	
	void queryId(uint id)
	{
		m_packet.queryId = id;
	}
	
	char[] keyword()
	{
		return mbsToUTF8(cast(ubyte[])getCString(m_packet.keyword));
	}
	
	char[] trip()
	{
		return getCString(m_packet.trip);
	}
	
	uint queryId()
	{
		return m_packet.queryId;
	}
}


// Command13
// クエリ送信

struct ViaAddress {
align(1):
	uint	address; 	// network byteoder
	short	port;		// x86 host byteorder
}
class ViaNode {
public:
	this(uint addr, ushort prt) 
	{
		address	= addr;
		port	= prt;
	}
	
	uint		address;
	ushort		port;
	
	char[] toString()
	{
		return (new InternetAddress(address, port)).toString;
	}
}

class NyQuery: NyCommand
{
public:
	enum Qr: ubyte {
		QUERY = 0,
		RESPONSE = 1
	}
protected:
	struct PackStructH {
	align(1):
		ubyte	qr;
		ubyte	diffusion_flag;
		ubyte	downstream_flag;
		ubyte	bbs_flag;
		uint	query_id;
		ubyte	keyword_length;
	};
//	ubyte		via_count;

//	ushort		key_count
	
	struct Trip {
		ubyte[11] trip;
	}
	Qr					m_qr;
	bool				m_is_diffusion_query;
	bool				m_is_downstream_query;
	bool				m_is_bbs_query;
	uint				m_query_id;
	char[]				m_keyword;
	char[]				m_trip;
	ViaNode[]			m_vianode;
	NyKeyInformation[]	m_keyinfo;
	
	void unpackInternal(ubyte[] data)
	{
		PackStructH *pack_h;
		int offset;
		if (commandCode != NyCommandCode.QUERY) {
			throw new NyBadCommand("Query: クエリでない");
		}
		
		if (data.length < PackStructH.sizeof) {
			throw new NyBadCommand("Query: データ不足");
		}
		
		// ヘッダ
		pack_h 	= cast(PackStructH *)data.ptr;
		data	= data[PackStructH.sizeof .. length];
		
		if (pack_h.qr != Qr.QUERY && pack_h.qr != Qr.RESPONSE) {
			throw new NyBadCommand("Query: Q/R範囲外");
		}
		m_qr					= cast(Qr)pack_h.qr;
		m_is_diffusion_query	= pack_h.diffusion_flag == 0 ? false: true;
		m_is_downstream_query	= pack_h.downstream_flag == 0 ? false: true;
		m_is_bbs_query			= pack_h.bbs_flag == 0 ? false: true;
		m_query_id				= pack_h.query_id;
		
		// 検索キーワード
		if (data.length < pack_h.keyword_length) {
			throw new NyBadCommand("Query: キーワード長異常");
		}
		m_keyword 	= mbsToUTF8(data[0 .. pack_h.keyword_length].dup);
		data		= data[pack_h.keyword_length .. length];
		
		// 検索トリップ
		if (data.length < Trip.sizeof) {
			throw new NyBadCommand("Query: トリップ異常");
		}
		m_trip = getCString(data[0 .. m_trip.sizeof].dup);
		data	= data[Trip.sizeof .. length];
		
		// 経由情報
		if (data.length < ubyte.sizeof) {
			throw new NyBadCommand("Query: 経由ノード情報リスト数");
		}
		ubyte via_count	= *(cast(ushort *)data.ptr);
		data				= data[ubyte.sizeof .. length];
		
		m_vianode = new ViaNode[via_count];
		for (ubyte i = 0; i < via_count; ++i) {
			if (data.length < ViaAddress.sizeof) {
				throw new NyBadCommand("Query: 経由ノード異常");
			}
			ViaAddress *addr 	= cast(ViaAddress *)data.ptr;
			m_vianode[i]		= new ViaNode(hostToNetwork32(addr.address), addr.port);
			data				= data[ViaAddress.sizeof .. length];
		}
		
		// キー情報
		if (data.length < ushort.sizeof) {
			throw new NyBadCommand("Query: キー情報リスト個数");
		}
		ushort key_count 	= *(cast(ushort *)data.ptr);
		data				= data[ushort.sizeof .. length];
		
		NyKeyInformation 	key;
		m_keyinfo = new NyKeyInformation[key_count];
		for (ushort j = 0; j < key_count; ++j) {
			key = new NyKeyInformation;
			key.unpack(data);
			m_keyinfo[j] = key;
			key = null;
		}
		data = null;
		
		clearBuffer();
	}
	
	uint getQueryID()
	{
		static uint id = 0;
		
		return ++id;
	}
	static int s_count;
public:
	this() {m_data = null;}
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	
	char [] toString()
	{
		char[] s =
			std.string.format(	"QR: %s, DIFFUSION: %s, DOWNSTREAM: %s, BBS: %s\n"
								"QUERY ID: %d, KEYWORD: %s\n"
								"TRIP: %s\n"
								"ViaNode.length: %s, Key.length: %s\n",
				cast(ubyte)m_qr, m_is_diffusion_query, m_is_downstream_query, m_is_bbs_query,
				m_query_id, m_keyword, m_trip, m_vianode.length, m_keyinfo.length);
		foreach (ViaNode n; m_vianode) {
			s ~= std.string.format("	- %s\n", n.toString);
		}
		
		return s;
	}
	
	char[] toStringFull()
	{
		char[] s =
			std.string.format(	"QR: %s, DIFFUSION: %s, DOWNSTREAM: %s, BBS: %s\n"
								"QUERY ID: %d, KEYWORD: %s\n"
								"TRIP: %s\n"
								"ViaNode.length: %s\n, Key.length: %s\n",
				cast(ubyte)m_qr, m_is_diffusion_query, m_is_downstream_query, m_is_bbs_query,
				m_query_id, m_keyword, m_trip, m_vianode.length, m_keyinfo.length);
		
		foreach (ViaNode n; m_vianode) {
			s ~= n.toString;
		}
		foreach (NyKeyInformation k; m_keyinfo) {
			s ~= k.toString;
		}
		
		return s;
	}
	
	// get	void unpack(NyRawCommand raw_command)
	override void unpack (NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal(m_data);
//		delete m_data;
//		m_data = null;
	}
	
	override void unpack(ubyte[] command_block)
	{
		ubyte[] data = super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal(data);
		
//		writef("Query - unpack %s\n", s_count++);
	}
	
	// set
	override ubyte[] pack()
	{
		uint		size, pos;
		PackStructH pack;
		ubyte		viacount;
		ubyte[11]	tripz;
		ViaAddress	vianode;
		ushort		keycount;
		ubyte[]		packet;
		
		ubyte[] mbs_keyword = utf8ToMBS(m_keyword);
		
		// ヘッダ
		pack.qr 				= m_qr;
		pack.diffusion_flag		= m_is_diffusion_query ? 1:0;
		pack.downstream_flag	= m_is_downstream_query ? 1:0;
		pack.bbs_flag			= m_is_bbs_query ? 1:0;
		pack.query_id			= m_query_id;
		pack.keyword_length		= mbs_keyword.length;
		
		// データ長計算
		size = PackStructH.sizeof + mbs_keyword.length + tripz.length 
				+ (m_vianode.length * ViaAddress.sizeof) + viacount.sizeof + keycount.sizeof;
		
		foreach (NyKeyInformation e; m_keyinfo) {
			size += e.length;
		}
		packet = new ubyte[CommandLength.sizeof + NyCommandCode.sizeof + size];
		
		// ベースのpackを呼ばないためここでコマンドヘッダも積める
		commandCode = NyCommandCode.QUERY;
		
		// length
		packet[0 .. CommandLength.sizeof] = uIntToUByteArray(NyCommandCode.sizeof + size);
		pos += CommandLength.sizeof;
		
		// command code
		packet[pos .. pos + NyCommandCode.sizeof] = m_command_code;
		pos += NyCommandCode.sizeof;
		
		// 以下クエリコマンド部
		
		packet[pos .. pos + PackStructH.sizeof] = (cast(ubyte *)&pack)[0 .. PackStructH.sizeof];
		pos += PackStructH.sizeof;
		
		packet[pos .. pos + mbs_keyword.length] = mbs_keyword;
		pos += mbs_keyword.length;
		
		// トリップ
		tripz[0 .. m_trip.length] = cast(ubyte[])m_trip[0 .. length];
		packet[pos .. pos + tripz.length] = tripz;
		pos += tripz.length;
		
		// 経由ノード
		viacount	= m_vianode.length;
		packet[pos .. pos + viacount.sizeof] = viacount;
		pos += viacount.sizeof;
		
		foreach (ViaNode node; m_vianode) {
			vianode.address		= networkToHost32(node.address);
			vianode.port		= node.port;
			packet[pos .. pos + ViaAddress.sizeof] = (cast(ubyte *)&vianode)[0 .. ViaAddress.sizeof];
			pos += ViaAddress.sizeof;
		}
		
		// キー情報
		keycount	= m_keyinfo.length;
		packet[pos .. pos + keycount.sizeof] = (cast(ubyte *)&keycount)[0 .. keycount.sizeof];
		pos += keycount.sizeof;
		
		foreach (NyKeyInformation key; m_keyinfo) {
			int len = key.length;
			packet[pos .. pos + len] = key.pack;
			pos += len;
		}
		m_data = null;
		
//		writef("Query -pack %s\n", s_count++);
		
		return packet;
	}
	//getter
	Qr qr()
	{
		return m_qr;
	}
	
	bool isDiffusionQuery()
	{
		return m_is_diffusion_query;
	}
	
	bool isDownstreamQuery()
	{
		return m_is_downstream_query;
	}
	
	bool isBBSQuery()
	{
		return m_is_bbs_query;
	}
	
	uint queryId()
	{
		return m_query_id;
	}
	
	char[] keyword()
	{
		return m_keyword;
	}
	
	char[] trip()
	{
		return m_trip;
	}
	
	ViaNode[] viaNode()
	{
		return m_vianode;
	}
	
	NyKeyInformation[] keyInformation()
	{
		return m_keyinfo;
	}
	
	// setter
	void qr(Qr qr_v)
	{
		m_qr = qr_v;
	}
	
	void isDiffusionQuery(bool v)
	{
		m_is_diffusion_query = v;
	}
	
	void isDownstreamQuery(bool v)
	{
		m_is_downstream_query = v;
	}
	
	void isBBSQuery(bool v)
	{
		m_is_bbs_query = v;
	}
	
	void queryId(ushort id)
	{
		m_query_id = id;
	}
	
	void keyword(char[] keyword)
	{
		m_keyword = keyword;
	}
	
	void trip(char[] trip)
	{
		m_trip = trip;
	}
	
	void viaNode(ViaNode[] node)
	{
		m_vianode = node;
	}
	
	void addViaNode(ViaNode node)
	{
		m_vianode ~= node;
	}
	
	void keyInformation(NyKeyInformation[] keyinfo)
	{
		m_keyinfo = keyinfo;
	}
	
	void addKeyInformation(NyKeyInformation keyinfo)
	{
		m_keyinfo ~= keyinfo;
	}
	
	override void destroy()
	{
//		writef("Query- destroy %s\n", s_count--);
		foreach (inout NyKeyInformation e; m_keyinfo) {
			e.destroy;
			delete e;
			e = null;
		}
		delete m_keyinfo;
		delete m_vianode;
		delete m_keyword;
		delete m_trip;
		m_keyinfo = null;
		m_vianode = null;
		m_keyword = null;
		m_trip = null;
	}
}

// Command11
// ファイル送信要求
class NyFileRequest: NyCommand
{
protected:
	struct PackStruct {
	align(1):
		uint 		task_id;
		uint 		begin_block;
		uint 		request_count;
		ubyte[16]	hash;
		uint		file_size;
	}
	
	uint 		m_task_id;
	uint 		m_begin_block;
	uint 		m_request_count;
	ubyte[16]	m_hash;
	uint		m_file_size;
	
	void unpackInternal(ubyte[] data)
	{
		if (data.length < PackStruct.sizeof) {
			throw new NyBadCommand("FileRequest: command is too small");
		}
		PackStruct *pack = cast(PackStruct *)data.ptr;
		
		m_task_id			= pack.task_id;
		m_begin_block 		= pack.begin_block;
		m_request_count		= pack.request_count;
		m_hash[0 .. length]	= pack.hash[0 .. length];
		m_file_size 		= pack.file_size;
		
//		delete m_data;
		m_data = null;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal(raw_command.data);
	}
	
	override void unpack(ubyte[] command_block)
	{
		ubyte[] data = super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal(data);
	}
	
	override ubyte[] pack()
	{
		PackStruct pack;
		
		pack.task_id			= m_task_id;
		pack.begin_block 		= m_begin_block;
		pack.request_count		= m_request_count;
		pack.hash[0 .. length]	= m_hash[0 .. length];
		pack.file_size 			= m_file_size;
		
		m_data = (cast(ubyte *)&pack)[0 .. PackStruct.sizeof].dup;
		
		commandCode	= NyCommandCode.FILE_REQUEST;
		
		return super.pack();
	}
	// getter
	
	uint taskId()
	{
		return m_task_id;
	}
	
	uint beginBlock()
	{
		return m_begin_block;
	}
	
	uint requestCount()
	{
		return m_request_count;
	}
	
	ubyte[] hash()
	{
		return m_hash.dup;
	}
	
	uint fileSize()
	{
		return m_file_size;
	}
	
	// setter
	
	void taskId(uint id)
	{
		m_task_id = id;
	}
	
	void beginBlock(uint block)
	{
		m_begin_block = block;
	}
	
	void requestCount(uint count)
	{
		m_request_count = count;
	}
	
	void hash(ubyte[] hash)
	{
		m_hash[0 .. length] = hash[0 .. m_hash.length];
	}
	
	void fileSize(uint file_size)
	{
		m_file_size = file_size;
	}
}

// Command21
// ファイル送信応答
class NyFileResponse: NyCommand
{
protected:
	struct PackStruct {
	align(1):
		uint 			task_id;
		uint 			begin_block;
		ubyte[16]		hash;
		ubyte[0x10000]	data;
	}
	PackStruct 	m_block;
	uint		m_data_size;
	
	void unpackInternal(ubyte[] data)
	{
		if (data.length 
			< m_block.task_id.sizeof + m_block.begin_block.sizeof + m_block.hash.sizeof) 
		{
			throw new NyBadCommand("NyResponse: command is too small");
		}
		m_data_size 
			= data.length - (m_block.task_id.sizeof + m_block.begin_block.sizeof + m_block.hash.sizeof);
		if (m_data_size > m_block.data.sizeof) {
			m_data_size = m_block.data.sizeof;
		}
		
		(cast(ubyte*)&m_block)[0 .. m_block.sizeof - (m_block.data.sizeof - m_data_size)] 
			= data[0 .. m_block.sizeof - (m_block.data.sizeof - m_data_size)];
		
		clearBuffer();
//		delete m_data;
//		delete data;
		m_data = null;
	}
	
public:
	const ubyte[] HEADER_KEY = [0x68, 0x65, 0x61, 0x64, 0x65, 0x72];
	
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal(raw_command.data);
	}
	
	override void unpack(ubyte[] command_block)
	{
		ubyte[] data = super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal(data);
	}
	
	override ubyte[] pack()
	{
		PackStruct pack;
		
		m_data = (cast(ubyte *)&m_block)[0 .. m_block.sizeof - (m_block.data.sizeof - m_data_size)].dup;
		
		commandCode	= NyCommandCode.FILE_RESPONSE;
		
		return super.pack();
	}
	
	// getter
	
	uint taskId()
	{
		return m_block.task_id;
	}
	
	uint beginBlock()
	{
		return m_block.begin_block;
	}
	
	ubyte[] hash()
	{
		return m_block.hash.dup;
	}
	
	ubyte[] fileData()
	{
		return m_block.data[0 .. m_data_size]; // do not dup
	}
	
	uint dataSize()
	{
		return m_data_size;
	}
	
	ubyte[] cacheBlock()
	{
		// 全部返す: 最終ブロックについてはゴミが付いているが変換側で対応する
		return (cast(ubyte *)&m_block)[uint.sizeof .. m_block.sizeof];
	}
	
	// setter
	void taskId(uint id)
	{
		m_block.task_id = id;
	}
	
	void beginBlock(uint block)
	{
		m_block.begin_block = block;
	}
	
	void hash(ubyte[] hash)
	{
		m_block.hash[0 .. length] = hash[0 .. m_block.hash.length];
	}
	
	void fileData(ubyte[] data)
	{
		uint endpos;
		
		if (data.length >  m_block.data.sizeof) {
			endpos = m_block.data.sizeof;
		} else {
			endpos = data.length;
		}
		m_data_size = endpos;
		m_block.data[0 .. endpos] = data[0 .. endpos]; // deep copy
	}
}

// Command31
// コネクション切断要求
class NyClose: NyCommand
{
protected:
	void unpackInternal()
	{
//		delete m_data;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal();
	}
	
	override void unpack(ubyte[] command_block)
	{
		super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal();
	}
	
	override ubyte[] pack()
	{
		commandCode	= NyCommandCode.CLOSE;
		m_data		= null;
		
		return super.pack();
	}
}

// Command31
// コネクション切断要求・接続制限
class NyConnectedLimitation: NyCommand
{
protected:
	void unpackInternal()
	{
//		delete m_data;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal();
	}
	
	override void unpack(ubyte[] command_block)
	{
		super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal();
	}
	
	override ubyte[] pack()
	{
		commandCode	= NyCommandCode.CONNECTED_LIMITATION;
		m_data		= null;
		
		return super.pack();
	}
}

// Command
// コネクション切断要求・リスナー接続不可
class NyWrongListeningPort: NyCommand
{

protected:
	void unpackInternal()
	{
//		delete m_data;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal();
	}
	
	override void unpack(ubyte[] command_block)
	{
		super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal();
	}
	
	override ubyte[] pack()
	{
		commandCode	= NyCommandCode.WRONG_LISTENING_PORT;
		m_data		= null;
		
		return super.pack();
	}
}


// Command34
// コネクション切断要求・無視ノード警告
class NyReject: NyCommand
{
protected:
	void unpackInternal()
	{
//		delete m_data;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal();
	}
	
	override void unpack(ubyte[] command_block)
	{
		super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal();
	}
	
	override ubyte[] pack()
	{
		commandCode	= NyCommandCode.REJECT;
		m_data		= null;
		
		return super.pack();
	}
}

// Command35
// コネクション切断要求・低速ノード警告
class NySlowRate: NyCommand
{
protected:
	void unpackInternal()
	{
//		delete m_data ;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal();
	}
	
	override void unpack(ubyte[] command_block)
	{
		super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal();
	}
	
	override ubyte[] pack()
	{
		commandCode	= NyCommandCode.SLOW_RATE;
		m_data		= null;
		
		return super.pack();
	}
}


// Command36
// コネクション切断要求・捏造警告
class NyLiar: NyCommand
{
protected:
	void unpackInternal()
	{
//		delete m_data;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal();
	}
	
	override void unpack(ubyte[] command_block)
	{
		super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal();
	}
	
	override ubyte[] pack()
	{
		commandCode	= NyCommandCode.LIAR;
		m_data		= null;
		
		return super.pack();
	}
}


// Command97
// コネクション切断要求・旧バージョン
class NyLowVersion: NyCommand
{
protected:
	void unpackInternal()
	{
//		delete m_data;
		clearBuffer();
	}
	
public:
	this(NyRawCommand raw_command)
	{
		unpack(raw_command);
	}
	this(){}
	
	void unpack(NyRawCommand raw_command)
	{
		assign(raw_command);
		unpackInternal();
	}
	
	override void unpack(ubyte[] command_block)
	{
		super.unpack(command_block, UnpackHow.HEADER_ONLY);
		unpackInternal();
	}
	
	override ubyte[] pack()
	{
		commandCode	= NyCommandCode.LOW_VERSION;
		m_data		= null;
		
		return super.pack();
	}
}
