﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.system.locker;

// 
// 無駄なくらいロックするマルチスレッド対応ワンコインロッカー
// 
// 
// try-finallyは、
// 代入演算子や比較演算子がオーバーロードされていて
// 例外も考えられるかナーといちおういれておいた
// 
// signal/releaseは
// 長いループ等で明示的にロックしたいときに使う
// ロッカーを開けっ放しでゴソゴソする感じ
// 
// signalはスレッドIDベースのチェックしかしていないので
// IDが短い期間で再利用されるような環境ではデバッグに注意
// 
// ソートはvaluesで取り出したやつをソートすべし

private import	std.thread;
private import	pes.system.semaphore;
private import 	std.stdio;
private import	pes.hashtable;

debug(locker)
{
	import std.stdio;
}

enum LockerLock: bool
{
	BLOCK,
	NON_BLOCK
}

class Locker(KT, VT)
{
public:
	enum Lock: LockerLock
	{
		BLOCK 		= LockerLock.BLOCK,
		NON_BLOCK	= LockerLock.NON_BLOCK
	}
	
protected:
	VT[KT]	m_aa;
	SemaphoreResource 	m_semaphore;
	semaphore			m_id;
//	HashTable!(KT, VT)  m_aa;
	
public:
	this(SemaphoreResource resource, semaphore id)
	{
		m_semaphore	= resource;
		m_id		= id;
		m_aa		= null;
	}
	
	void signal()
	{
		m_semaphore.signal(m_id);
	}
	
	void release()
	{
		m_semaphore.release(m_id);
	}
	
	void set(KT key, VT value)
	{
		signal();
		try {
			m_aa[key] = value;
//			m_aa.set(key, value);
		} finally {
			release();
		}
	}
	
	VT get(KT key, Lock lock = Lock.BLOCK)
	{
		VT value;
		
		if (lock == Lock.BLOCK) {
			signal();
			try {
				value = m_aa[key];
//				value = m_aa.get(key);
			} finally {
				release();
			}
		} else {
			value = m_aa[key];
//			value = m_aa.get(key);
		}
		
		return value;
	}
	
	// これは取得中だけしかロックしないので操作中に不整合が起こりうる
	VT[] values(Lock lock = Lock.BLOCK)
	{
		VT[] values;
		
		if (lock == Lock.BLOCK) {
			signal();
			try {
				// aaの実装上valuesはコピーされた配列が戻るため
				// ここではコピーを作らない
				values = m_aa.values;
			} finally {
				release();
			}
		} else {
			values = m_aa.values;
		}
		
		return values;
	}
	
	KT[] keys(Lock lock = Lock.BLOCK)
	{
		KT[] keys;
		
		if (lock == Lock.BLOCK) {
			signal();
			try {
				keys = m_aa.keys;
			} finally {
				release();
			}
		} else {
			keys = m_aa.keys;
		}
		
		return keys;
	}
	
	uint length(Lock lock = Lock.BLOCK)
	{
		uint length;
		
		if (lock == Lock.BLOCK) {
			signal();
			try {
				length = m_aa.length;
			} finally {
				release();
			}
		} else {
			length = m_aa.length;
		}
		
		return length;
	}
	
	bool exists(KT key, Lock lock = Lock.BLOCK)
	{
		bool found = false;
		
		if (lock == Lock.BLOCK) {
			signal();
			try {
				if (key in m_aa) {
					found = true;
				}
			} finally {
				release();
			}
		} else {
			if (key in m_aa) {
				found = true;
			}
		}
		
		return found;
	}
	
	void remove(KT key)
	{
		signal();
		try {
			if (key in m_aa) {
				m_aa.remove(key);
			}
		} finally {
			release();
		}
	}
	
	void clear()
	{
		signal();
		try {
			foreach (KT e; m_aa.keys) {
				remove(e);
			}
			m_aa = null;
		} finally {
			release();
		}
	}
	
	bool empty(Lock lock = Lock.BLOCK)
	{
		bool is_empty;
		
		if (lock == Lock.BLOCK) {
			signal();
			try {
				is_empty = m_aa.length == 0 ? true: false;
			} finally {
				release();
			}
		} else {
			is_empty = m_aa.length == 0 ? true: false;
		}
		
		return is_empty;
	}
	
	void rehash()
	{
		signal();
		try {
			m_aa.rehash;
		} finally {
			release();
		}
	}
}

