﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module worker.cachedecoder;

// キャッシュデコーダー

private import pes.data.sqlite3;
private import std.md5;
private import std.file;
private import std.path;
private import std.thread;
private import std.base64;
private import std.c.time;
private import std.mmfile;
private import std.stream;
private import std.gc;
private import std.random;
private import std.c.stdlib;


private import pes.checksum;
private import pes.rc4;
private import pes.file;
private import pes.conv;
private import pes.data.sqlite3;
private import pes.system.all;

private import worker.cachebitmap;
private import common.workertable;
private import common.environment;

private const char[] CACHE_BLOCK_SELECT =
	"SELECT * FROM cache_block_table ORDER BY no";

class CacheDecoder: WorkerTask
{
protected:
	ulong			m_size;
	ubyte[] 		m_hash;
	Environment 	s_environ;
	int				m_result;
	char[]			m_errmsg;
	char[]			m_name_output;
	
	const ubyte[] CACHE_KEY_INIT = [0x68, 0x65, 0x61, 0x64, 0x65, 0x72];
	
	
	int mainLoop()
	{
//		setPriority(PRIORITY.DECREASE);
//		setPriority(PRIORITY.INCREASE);
		gcLimit = 16;
		
		try {
			m_result = threadMain();
		} catch (Exception e) {
			m_errmsg = e.toString;
			m_result = -1;
		}
		
		std.gc.fullCollect();
		
		return m_result;
	}
	
	int threadMain()
	{
		char[] 		filepath 	= filePath(m_hash);
		CacheHeader	header;
		
		if (!std.file.exists(filepath)) {
			return -1; // 存在しないため終了
		}
		header	= cacheHeader();
		name	= header.fileName;
		
		decode(filepath, header);
		
		//delete header;
		header = null;
		
		return 0;
	}
	
	void toValidPath(char[] path)
	{
		for (int i = 0; i < path.length; ++i) {
			if (path[i] == '=') {
				path[i] = '_';
			} else if (path[i] == '/') {
				path[i] = '-';
			} else if (path[i] == '+') {
				path[i] = ' ';
			}
		}
	}
	
	CacheHeader cacheHeader()
	{
		CacheHeader header 	= s_environ.cacheHeader.get(m_hash, m_size);
		m_name_output = null;
		
		if (header.fileName.length == 0) {
			m_name_output = std.string.toString(std.random.rand()) ~ ".nazo";
		}
		if (header.completed) {
			m_name_output 	= s_environ.configuration.downloadFolder 
									~ std.path.sep ~ header.fileName;
			
			// 同じファイル名がある場合
			if (std.file.exists(m_name_output)) {
				if (! pes.file.isDifferent(m_name_output, header.hash)) {
					// 既にデコード済みで中断
					throw new WorkerTermianteEvent;
				}
				// 同じファイル名で違うファイルはハッシュ値をつける
				char[] name = std.string.format("%s_%s.%s",
								getName(m_name_output),
								binaryToHexString(header.hash), getExt(m_name_output));
				if (std.file.exists(name)) {
					// それでも存在する
					if (! pes.file.isDifferent(name, header.hash)) {
						// 既にデコード済みで中断
						throw new WorkerTermianteEvent;
					}
					// 上書き
				}
				
				m_name_output = name;
			}
			try {
				char[] incomplete_name = s_environ.configuration.downloadFolder 
											~ std.path.sep ~ "_incomplete_" ~ header.fileName;
				if (std.file.exists(incomplete_name)) {
					forcedRemoveFile(incomplete_name);
				}
			} catch (Exception) {
				// do nothing
			}
		} else {
			m_name_output	= s_environ.configuration.downloadFolder 
									~ std.path.sep ~ "_incomplete_" ~ header.fileName;
		}
		
		return header;
	}
	
	
	char[] filePath(ubyte[] hash)
	{
		uint	blen		= std.base64.encodeLength(hash.length);
		char[]	name		= new char[blen];
		
		name = std.base64.encode(cast(char[])hash, name);
		toValidPath(name);
		
		return s_environ.configuration.cacheFolder ~ std.path.sep ~ name ~ ".db";
	}
	
	void decode(char[] cache, CacheHeader header)
	{
		char[]			tempname	= m_name_output ~ ".swp";
		RC4				rc4ctx		= new RC4;
		File			stream		= new File(tempname, FileMode.OutNew);
		SQLite			db			= new SQLite(cache);
		SQLiteStatement stmt;
		MD5_CTX 		context;
		ubyte[16]		allsum;
		ulong			block_count = 0;
		ulong			data_size;
		ubyte[]			key			= CACHE_KEY_INIT;
		uint			pos = 0;
		
		ubyte[]			sum;
		ulong			no;
		ubyte[]			block;
		ubyte* 			buffer;
		
		// キャッシュファイル作成
		
		db.begin();
		uint select_block = 64;
		uint block_size = 0xffff;
		
		workMax(cast(uint)header.dataSize);
		try {
			
			context.start();
			data_size = 0;
			while (1) {
				bool block_found = false;
				buffer = cast(ubyte *)malloc((select_block + 1) * block_size);
				int buffer_size = 0;
				
				stmt = db.prepare(
					std.string.format(
						"SELECT no, hash, block FROM cache_block_table "
						"WHERE no >= %d AND no < %d "
						"ORDER BY no", block_count, block_count + select_block
					)
				);
				for (stmt.step(); stmt.hasRow; stmt.step()) {
					block_found = true;
					pos = 0;
					no			= stmt.fetch!(ulong)(pos++);
					sum			= stmt.fetch!(ubyte[])(pos++);
					block 		= stmt.fetch!(ubyte[])(pos++);
					
					if (no != block_count) {
						log(format("end of block no %s", block_count));
						break; // ブロックずれ
					}
					
					if (m_size < data_size + block.length) {
						ulong diff = data_size + block.length;
						
						diff -= m_size;
						
						block = block[0 .. length - diff];
					}
					data_size += block.length;
					
					rc4ctx.key(key);
					rc4ctx.crypt(block);
					if (!checkMD5Sum(sum, block)) {
						log("md5 sum error");
						break; // サムエラー
					}
					if (! header.bitmap.isSet(block_count)) {
						header.bitmap.set(block_count); /// 整合性復元処理
					}
					
					context.update(block);
					//stream.writeBlock(block.ptr, block.length);
					// 
					buffer[buffer_size .. buffer_size + block.length] = block;
					buffer_size += block.length;
					
					++block_count;
					key = cast(ubyte[])std.string.format("%d %s", block_count, binaryToHexString(m_hash));
					
					work(block.length); // 作業量更新
					
					pollTerminateEvent(); // 中断チェック
	//				this.fullCollect();
					delayTask(10);
				}
				
				if (!block_found) {
					free(buffer);
					buffer = null;
					stmt.finalize();
					break;
				}
				stream.writeBlock(buffer, buffer_size);
				free(buffer);
				buffer = null;
				stmt.finalize();
			}
		} finally {
			if (buffer != null) {
				free(buffer);
			}
			stream.close();
//			stmt.finalize();
			db.end();
			db.close();
			s_environ.cacheHeader.update(header);
			if (std.file.exists(m_name_output)) {
				forcedRemoveFile(m_name_output);
			}
			std.file.rename(tempname, m_name_output);
		}
		
		context.finish(allsum);
		if (header.completed 
			&& allsum != header.hash) 
		{
			// 全体SUM不一致
//			s_environ.cacheHeader.remove(header.hash, header.dataSize);
//			forcedRemoveFile(cache);
			log("md5 sum error occur in completed file");
			m_result = -2;
		}
	}
	
public:
	this(TaskResource resource, ubyte[] hash, uint data_size, Environment environ)
	{
		super(ANONYMOUS_MAIL, resource);
		m_hash 		= hash;
		m_size		= data_size;
		s_environ	= environ;
	}
	
	int exitState()
	{
		return m_result;
	}
	
	char[] errorMessage()
	{
		return m_errmsg;
	}
	
	ubyte[] hash()
	{
		return m_hash;
	}
}
