﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module worker.uploader;

// アップローダー

private import std.socket;
private import std.stdio;
private import std.path;
private import std.file;
private import std.thread;
private import std.random;
private import std.gc;

private import pes.net.winny.connection;
private import pes.net.winny.command;
private import pes.conv;
private import pes.rc4;
private import pes.checksum;
private import pes.data.sqlite3;
private import pes.system.all;

private import pes.dump;

private import common.cacheresolve;
private import common.task;
private import common.environment;
private import common.workertable;

private const char[] CACHE_BLOCK_SELECT =
	"SELECT * FROM cache_block_table WHERE no BETWEEN ? AND ? ORDER BY no";

debug(download)
{
	import pes.prompt;
}

class UploadFile {
	uint	taskId;
	ubyte[]	hash;
	uint	dataSize;
	uint	beginBlock;
	uint	count;
	
	this(uint task_id, ubyte[] hash_md5, uint data_size, uint begin_block, uint count)
	{
		taskId		= task_id;
		hash 		= hash_md5;
		dataSize	= data_size;
		beginBlock	= begin_block;
		this.count	= count;
	}
}

class Uploader: WorkerTask
{
public:
	
protected:
	Connection 		m_connection;
	RequestStatus	m_status;
	Environment		s_environ;
	UploadFile		m_request;
	
	int mainLoop()
	{
		NyClose	close = new NyClose;
		char[] 	cache = cacheResolve(m_request.hash, s_environ.configuration);
		
		gcLimit = 16;
		
		if (!std.file.isfile(cache)) {
			debug(download) mbswritef(derr, "Uploader: キャッシュなし %s %s-%s\n",
			m_request.hash, m_request.beginBlock, m_request.beginBlock + m_request.count);
			
			m_status = RequestStatus.NG;
			try {
				m_connection.connection.sendCommand(close);
			} finally {
				m_connection.unlock();
			}
			
			return 0;
		}
		
		name = binaryToHexString(m_request.hash);
		
		try {
			debug(download) mbswritef(derr, "Uploader: アップロード開始 %s %s-%s\n",
				m_request.hash, m_request.beginBlock, m_request.beginBlock + m_request.count);
			
			// blocking
			m_connection.connection.blocking = true;
			
			upload(cache);
		}
		catch (Object e) {
			debug(download) mbswritef(derr, "Uploader: Exception %s\n", e.toString);
			
			m_status = RequestStatus.NG;
			m_connection.close(); // 例外は切断する
		} finally {
			// remove
			m_connection.connection.blocking = false;
			m_connection.unlock();
//			debug(download) mbswritef(derr, "Uploader: コネクション削除\n");
//			s_environ.ulDataConnection.remove(m_connection.aaKey);
//			debug(download) mbswritef(derr, "Uploader: コネクション削除2\n");
		}
		
		std.gc.fullCollect();
		
		m_status = RequestStatus.OK;
		debug(download) mbswritef(derr, "Uploader: アップロード終了\n");
		
		return 0;
	}
	
	void upload(char []cache)
	{
		SQLite 			db				= new SQLite(cache);
		SQLiteStatement	stmt;
		NyFileResponse	response		= new NyFileResponse;
		uint	block_no				= m_request.beginBlock;
		uint	no, pos;
		
		try {
			workMax(m_request.count * 0x10000);
			
			stmt = db.prepare(CACHE_BLOCK_SELECT);
			stmt.bind(1, m_request.beginBlock, m_request.beginBlock + m_request.count -1);
			
			for (stmt.step(); stmt.hasRow; stmt.step()) {
				debug(download) mbswritef(derr, "Uploader: アップロード%s \n", block_no);
				
				pos = 0;
				no = stmt.fetch!(uint)(pos++);
				if (no != block_no) {
					debug(download) mbswritef(derr, "Uploader: ブロックとび\n");
					try {
						m_connection.connection.sendCommand(new NyClose);
					} catch (SocketException){}
					break;
				}
				
				response.taskId 	= m_request.taskId;
				response.beginBlock	= no;
				response.hash		= stmt.fetch!(ubyte[])(pos++);
				response.fileData	= stmt.fetch!(ubyte[])(pos++);
				
				try {
					pollTerminateEvent(); // 中断チェック
					
					m_connection.connection.sendCommand(response);
					
					pollTerminateEvent(); // 中断チェック
					
				} catch (SocketException e) {
					debug(download) mbswritef(derr, "Uploader: %s:%s\n", e.toString, e.errorCode);
					break;
				}
				
				++block_no;
				work(0x10000);
				
				this.fullCollect();
				delayTask(10);
			}
		} finally {
			delete response;
			response = null;
			if (stmt !is null) {
				stmt.finalize();
			}
			db.close();
		}
		s_environ.cacheHeader.referred(m_request.hash,
					m_request.dataSize, block_no - m_request.beginBlock);
	}
	
public:
	this(TaskResource resource, Connection connection,
			UploadFile request, Environment environ)
	{
		super(ANONYMOUS_MAIL, resource);
		m_connection 	= connection;
		m_request		= request;
		s_environ		= environ;
		m_status		= RequestStatus.NG;
	}
	
	ubyte[] hash()
	{
		return m_request.hash;
	}
	
	RequestStatus status()
	{
		return m_status;
	}
	
	// 上がり通信レート最大値設定
	void rate(uint max)
	{
		s_environ.ulDataConnection.signal();
		try {
			if (m_connection !is null 
				&& m_connection.connection !is null) 
			{
				m_connection.connection.upstreamRateMax = max;
			}
		} finally {
			s_environ.ulDataConnection.release();
		}
	}
	
	override uint rate()
	{
		if (m_connection !is null
			&& m_connection.connection !is null) 
		{
			return m_connection.connection.rate();
		}
		
		return super.rate();
	}
	
	UploadFile requestFile()
	{
		return m_request;
	}
}

